#include "stdafx.h"
#include "resource.h"
#include "UnicodeData.h"


/****************************************************************************
*                              UnicodeData::Load
* Result: BOOL
*       TRUE if Unicode data loaded successful
*       FALSE if error (use GetLastError())
* Effect: 
*       Loads the Unicode data and sets up a pointer to it; builds an index
****************************************************************************/

BOOL UnicodeData::Load()
    {
     HRSRC resource;
     resource = ::FindResource(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDR_UNICODE), _T("UNICODE"));
     if(resource == NULL)
        { /* failed */
         DWORD err = ::GetLastError();
        // TODO: Report error here
         return FALSE;
        } /* failed */

     HGLOBAL h = ::LoadResource(AfxGetInstanceHandle(), resource);
     if(h == NULL)
        { /* load failed */
         DWORD err = ::GetLastError();
        // TODO: Report error here
         return FALSE;
        } /* load failed */

     data = (LPCSTR)::LockResource(h); // note this is an 8-bit resource!
     if(data == NULL)
        return FALSE;

     size = SizeofResource(AfxGetInstanceHandle(), resource);
     
     UINT last = 0xFFFFFFFF;
     LPCSTR p = data;
     while(TRUE)
        { /* scan data */
         if(p >= data + size)
            break; // end of data
         if(*p == '#' || *p == '\r') // 8-bit test
            { /* next line */
             p = strchr(p, '\n'); // 8-bit test
             if(p == NULL)
                break; // done
             p++; // start of next line
             continue;
            } /* next line */

         UINT codepoint = strtoul(p, NULL, 16);
         UINT group = codepoint >> 8;
         if(group != last)
            { /* changed */
             if(group < (UINT)Index.GetSize() && Index[group] != NULL)
                ASSERT(FALSE);
             Index.SetAtGrow(group, p);
             last = group;
             TRACE(_T("[%04xXX] %p\n"), group, p);
            } /* changed */
         p = strchr(p, '\n'); // 8-bit test
         if(p == NULL)
            break; // end of data
         p++;
        } /* scan data */
     return TRUE;
    } // UnicodeData::Load

/****************************************************************************
*                          UnicodeData::GetFirstCode
* Result: POSITION
*       Position of first code
*       NULL if error
* Effect: 
*       Marks the position of the first valid code
****************************************************************************/

POSITION UnicodeData::GetFirstCode()
    {
     LPCSTR p = data;
     if(p == NULL)
        return NULL;

     while(TRUE)
        { /* scan for first */
         if(p >= data + size)
            break; // end of data
         if(*p == '#' || *p == '\r') // 8-bit test
            { /* next line */
             p = strchr(p, '\n'); // 8-bit test
             if(p == NULL)
                break; // done
             p++; // start of next line
             continue;
            } /* next line */
         break; // got a valid line
        } /* scan for first */
     return (POSITION)p; // return pointer
    } // UnicodeData::GetFirstCode

/****************************************************************************
*                          UnicodeData::GetNextCode
* Inputs:
*       POSITION & p: Position value
* Result: LPCSTR
*       Pointer to string, or NULL if no more strings
* Effect: 
*       Updates p to be the next position
****************************************************************************/

LPCSTR UnicodeData::GetNextCode(POSITION & pos)
    {
     LPCSTR p = (LPCSTR)pos;
     if(p == NULL)
        return NULL;  // should not be here

     while(TRUE)
        { /* scan for next line */
         if(p >= data + size)
            { /* end of data */
             pos = (POSITION)0;
             return NULL;
            } /* end of data */

         if(*p == '#' || *p == '\r') // 8-bit test
            { /* next line */
             p = strchr(p, '\n'); // 8-bit test
             if(p == NULL)
                { /* end of data */
                 pos = (POSITION)0;
                 return NULL;
                } /* end of data */
             p++; // start of next line
             continue;
            } /* next line */
         break; // got a valid character
        } /* scan for next line */
     LPCSTR result = p;
     p = strchr(p, '\n');
     if(p == NULL)
        { /* end of data */
         pos = (POSITION)0;
        } /* end of data */
     else
        { /* more data */
         p++;
         pos = (POSITION)p;
        } /* more data */
     return result;
    } // UnicodeData::GetNextCode

/****************************************************************************
*                      UnicodeData::GetField
* Inputs:
*       UINT col: Field #
*       LPCSTR p: Pointer to string value
*               xxxx;xxxx;xxxx;xxxx;xxxx;;;;xxxx;xxxx;;;;
*               ^
*               |
*               p
* Result: UnicodeData::Substring
*       The Substring descriptor
****************************************************************************/

UnicodeData::Substring UnicodeData::GetField(UINT col, LPCSTR p)
    {
     if(p == NULL)
         return UnicodeData::Substring("", 0);
     while(TRUE)
        { /* scan cols */
         int len = strcspn(p, ";\r\n"); // 8-bit tests!
         if(col == 0)
            return Substring(p, len);
         // Not the desired column
         // 0000;NAMEHERE;xx;yy;;;;
         // ^   ^
         // |   |
         // p   p+len
         col--;  // decrement column count
         p += len + 1;
        } /* scan cols */
    } // UnicodeData::GetField

/****************************************************************************
*                          UnicodeData::GetNextField
* Inputs:
*       const Substring & ss: Substring
* Result: Substring
*       Pointer to the next field
****************************************************************************/

UnicodeData::Substring UnicodeData::GetNextField(const Substring & ss)
    {
     LPCSTR p = ss.GetPtr();
     p += ss.GetLength();
     p++;
     int n = strcspn(p, ";\r\n"); // 8-bit tests!
     return Substring(p, n);
    } // UnicodeData::GetNextField
