#include "stdafx.h"
#include "RTLMap.h"
#include "Registry.h"
#ifdef _DEBUG
#include <malloc.h> // for _heapchk
#endif

#ifdef _UNICODE
static const LCID RTL[] = {
   0x040d, // Hebrew
   0x0401, // Arabic (Saudi Arabia) 
   0x0801, // Arabic (Iraq) 
   0x0c01, // Arabic (Egypt) 
   0x1001, // Arabic (Libya) 
   0x1401, // Arabic (Algeria) 
   0x1801, // Arabic (Morocco) 
   0x1c01, // Arabic (Tunisia) 
   0x2001, // Arabic (Oman) 
   0x2401, // Arabic (Yemen) 
   0x2801, // Arabic (Syria) 
   0x2c01, // Arabic (Jordan) 
   0x3001, // Arabic (Lebanon) 
   0x3401, // Arabic (Kuwait) 
   0x3801, // Arabic (U.A.E.) 
   0x3c01, // Arabic (Bahrain) 
   0x4001, // Arabic (Qatar) 
   0x0429, // Farsi 
   0 // EOT
};
#endif

/****************************************************************************
*                          CLocaleExplorerDlg::IsRTL
* Inputs:
*       LCID lcid: Locale ID
* Result: BOOL
*       TRUE if the lcid is a right-to-left-reading locale
****************************************************************************/

#ifdef _UNICODE
static BOOL IsRTL(LCID lcid)
    {
     for(int i = 0; RTL[i] != 0; i++)
        { /* scan */
         if(RTL[i] == lcid)
            return TRUE;
        } /* scan */
     return FALSE;
    } // IsRTL
#endif // _UNICODE

/****************************************************************************
*                       CLocaleExplorerDlg::HasUnicode
* Inputs:
*       CString text: Text string to examine
* Result: BOOL
*       TRUE if there are any characters > U00FF in the string
*       FALSE if all characters are <= U00FF
****************************************************************************/

#ifdef _UNICODE
static BOOL HasUnicode(const CString & text)
    {
     for(int i = 0; i < text.GetLength(); i++)
        { /* check it out */
         WCHAR ch = text[i];
         if(ch > 0x00FF)
            return TRUE;
        } /* check it out */
     return FALSE;
    } // HasUnicode
#endif

/****************************************************************************
*                                     RTLMap
* Inputs:
*       CString & text: Text string
*       LCID lcid: Locale ID
* Result: void
*       
* Effect: 
*       If the string is Unicode, has a character > 0000..00FF, and is
*       in a right-to-left reading locale, swap the order
****************************************************************************/

#ifdef _UNICODE
void RTLMap(CString & text, LCID lcid)
    {
     if(!IsRTL(lcid))
        return; // already in correct order

     if(!HasUnicode(text))
        return;

     if(text.IsEmpty())
        return; // nothing to do with empty strings

     if(text.GetLength() == 1)
        return; // ignore 1-character strings
     
     // Now the problem arises that numbers need to remain in the same order
     // e.g.,                  ABCDEFG 123 HIJKLMN
     // should be changed to   NMLKJIH 123 GFEDCBA
     
     // state machine
     // state           char             newstate        computation
     //----------------------------------------------------------------
     // A. LTR          C2_LEFTTORIGHT  LTR             LTRdata += text[i]
     // B. LTR          C2_RIGHTTOLEFT  RTL             result = text[i] + LTRdata + result;
     // C. LTR          NUL             n/a             result = LTRdata + result;
     // D. RTL          C2_LEFTTORIGHT  LTR             LTRdata = text[i]
     // E. RTL          C2_RIGHTTOLEFT  RTL             result = text[i] + result
     // F. RTL          NUL             n/a             ------

     enum {LTR, RTL} direction = LTR;
     CString result;
     CString LTRdata;
     
     CWordArray chartypes2;
     chartypes2.SetSize(text.GetLength()); 
     LPWORD p2 = chartypes2.GetData();
     if(!::GetStringTypeEx(lcid, CT_CTYPE2, text, text.GetLength(), p2))
        { /* failed */
         return; // can't use info
        } /* failed */

#ifdef _DEBUG //----------------------------------------------------------------+
     CWordArray chartypes1;                                                  // |
     chartypes1.SetSize(text.GetLength());                                   // |
     LPWORD p1 = chartypes1.GetData();                                       // |
     if(!::GetStringTypeEx(lcid, CT_CTYPE1, text, text.GetLength(), p1))     // |
        { /* failed */                                                       // |
         return;                                                             // |
        } /* failed */                                                       // |
#endif //-----------------------------------------------------------------------+

     for(int i = 0; i < text.GetLength(); i++)
        { /* scan string */
         //ASSERT( (chartypes1[i] & C1_DIGIT) == 0);
         if(direction == LTR)
            { /* LTR */
             switch(chartypes2[i])
                { /* LTR chartypes2 */
                 case C2_LEFTTORIGHT:
                 case C2_ARABICNUMBER:
                 case C2_EUROPENUMBER:
                 case C2_EUROPESEPARATOR:
                 case C2_EUROPETERMINATOR:
                 case C2_COMMONSEPARATOR:
                 case C2_BLOCKSEPARATOR:
                 case C2_SEGMENTSEPARATOR:
                 case C2_WHITESPACE:
                 case C2_OTHERNEUTRAL:
                 case C2_NOTAPPLICABLE:
                    LTRdata += text[i];
                    break;
                 case C2_RIGHTTOLEFT:
                    result = text[i] + LTRdata + result;
                    direction = RTL;
                    break;
                } /* LTR chartypes2 */
            } /* LTR */
         else
            { /* RTL */
             switch(chartypes2[i])
                { /* RTL chartypes2 */
                 case C2_LEFTTORIGHT:
                 case C2_ARABICNUMBER:
                 case C2_EUROPENUMBER:
                 case C2_EUROPESEPARATOR:
                 case C2_EUROPETERMINATOR:
                 case C2_COMMONSEPARATOR:
                 case C2_BLOCKSEPARATOR:
                 case C2_SEGMENTSEPARATOR:
                 case C2_WHITESPACE:
                 case C2_OTHERNEUTRAL:
                 case C2_NOTAPPLICABLE:
                    LTRdata += text[i];
                    direction = LTR;
                    break;
                 case C2_RIGHTTOLEFT:
                    result = text[i] + result;
                    break;
                } /* RTL chartypes2 */
            } /* RTL */
#if 0                   
         if(direction == LTR && ((chartypes2[i] == C2_LEFTTORIGHT) || chartypes2[i] == C2_ARABICNUMBER))
            { /* A */
             LTRdata += text[i];
            } /* A */
         else
         if(direction == LTR && (chartypes2[i] == C2_RIGHTTOLEFT))
            { /* B */
             result = text[i] + LTRdata + result;
             direction = RTL;
            } /* B */
         else
         if(direction == RTL && ((chartypes2[i] == C2_LEFTTORIGHT || chartypes2[i] == C2_ARABICNUMBER)))
            { /* D */
             LTRdata = text[i];
             direction = LTR;
            } /* D */
         else
         if(direction == RTL && (chartypes2[i] == C2_RIGHTTOLEFT))
            { /* E */
             result = text[i] + result;
            } /* E */
         else
            ASSERT(FALSE);
#endif
        } /* scan string */

     if(direction == LTR)
        { /* C */
         result = LTRdata + result;
        } /* C */
     else
        { /* F */
         // does nothing
        } /* F */
     text = result;
    } // CLocaleExplorerDlg::RTLMap
#endif
